/*****************************************************************************

  3D Randomized Motion Planner

  COPYRIGHT (c) 1990 
  THE BOARD OF TRUSTEES OF THE LELAND STANFORD JUNIOR UNIVERSITY
  ALL RIGHTS RESERVED

  No duplications, whole or partial, manual or electronic, may be made 
  without the express permission of the author. Any such copies, or 
  revisions thereof, must display this notice unaltered.

  Robotics Laboratory,
  Computer Science Department,
  Stanford University

  Original author: Jerome Barraquand
  Modified by Tsai-Yen Li (li@flamingo.Stanford.EDU)
  Advisor: Jean-Claude Latombe

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <time.h>
#include "defs.h"
#include "solidstruct.h"
#include "macro.h"
#include "synth3dp.h"

extern FILE *log_fd;

/*****************************************************************************
  Function to compute the distance map and skeleton
****************************************************************************/

void  distance(descript ws,unsigned char ***ws_bmap,unsigned char ***V,
	       unsigned char ***voro, float grow){

  register int i,j,k,l,m,x,y,z;
  pixel *FRONT,*NEXT,*TEMP;
  int nbfront = 0, nbnext = 0 ;
  pixel ****pere,*pepere,*ppepere;
  int px,py,pz;
  int igrow;
  int dimx=ws.dimx,dimy=ws.dimy,dimz=ws.dimz;
  
  FRONT = (pixel *)calloc(10*(dimy*dimz),sizeof(pixel));
  NEXT  = (pixel *)calloc(10*(dimy*dimz),sizeof(pixel));
  ppepere = pepere = (pixel *)calloc(30*(dimy*dimz),sizeof(pixel));
  
  pere = (pixel ****)calloc3D(dimx,dimy,dimz,sizeof(pixel *));

  memcpy(&(V[0][0][0]),&(ws_bmap[0][0][0]),dimx*dimy*dimz);

  igrow=(int)(grow*dimx);

  /* grow the obstacle by the max. thickness of the robot */

#ifdef VERBOSE
  if (verbose)
    fprintf(stderr,"Number of growing boundary: %d\n",igrow);
#endif
#ifdef DEBUG
    fprintf(log_fd,"Number of growing boundary: %d\n",igrow);
#endif
  
  for(i=0;i<igrow;i++) {
    memcpy(&(ws_bmap[0][0][0]),&(V[0][0][0]),dimx*dimy*dimz);
    j=0;
    for(x=0;x<dimx;++x) {
      for(y=0;y<dimy;++y) {
	for(z=0;z<dimz;++z) {
	  if((ws_bmap[x][y][z] != OBSTACLE) && 
	     ((ws_bmap[x+1][y][z]==OBSTACLE)||(ws_bmap[x-1][y][z]==OBSTACLE)||
	      (ws_bmap[x][y+1][z]==OBSTACLE)||(ws_bmap[x][y-1][z]==OBSTACLE)|| 
	      (ws_bmap[x][y][z+1]==OBSTACLE)||(ws_bmap[x][y][z-1]==OBSTACLE))){
	    V[x][y][z] = OBSTACLE; 
	    j++;
	  }
	}
      }
    }
#ifdef VERBOSE
  if (verbose)
    fprintf(stderr,"Growing #%d\n",i);
#endif
#ifdef DEBUG
    fprintf(log_fd,"In %d growth, add %d cells to the bitmap.\n",i,j);
    if (verbose)
      fprintf(stderr,"In %d growth, add %d cells to the bitmap.\n",i,j);
#endif
  }

  /* find the base of the wavefront */

  for(x=0; x<dimx; ++x) {
    for(y=0; y<dimy; ++y) {
      for(z=0; z<dimz; ++z) {
	if(ws_bmap[x][y][z]!= OBSTACLE && 
	   (ws_bmap[x+1][y][z] == OBSTACLE || 
	    ws_bmap[x-1][y][z] == OBSTACLE ||
	    ws_bmap[x][y+1][z] == OBSTACLE || 
	    ws_bmap[x][y-1][z] == OBSTACLE || 
	    ws_bmap[x][y][z+1] == OBSTACLE || 
	    ws_bmap[x][y][z-1] == OBSTACLE)) { 
	  V[x][y][z] = 0; 
	  FRONT[nbfront].x = pepere->x = x;
	  FRONT[nbfront].y = pepere->y = y;
	  FRONT[nbfront].z = pepere->z = z; 
	  pere[x][y][z] = pepere++;
	  nbfront++;
	}
      }
    }
  }

#ifdef VERBOSE
  if (verbose)
    fprintf(stderr,"First wavefront found.\n");
#endif
  
  /* find the wavefront */

  for(i=0;i<nbfront;++i){

    x = FRONT[i].x;
    y = FRONT[i].y;
    z = FRONT[i].z;

    for(k= -1;k<=1;++k){
      for(l= -1;l<=1;++l){
	for(m= -1;m<=1;++m){
	  if(k == 0 && l == 0 && m == 0) continue;
	  if(ws_bmap[x+k][y+l][z+m] == OBSTACLE )
	    pere[x+k][y+l][z+m] = pere[x][y][z];
	}
      }
    }
  }

#ifdef VERBOSE
  if (verbose)
    fprintf(stderr,"Bases of wavefront found.\n");
#endif

  /* propagate the wavefront */
    
  k=0;
  while(nbfront > 0) {
    nbnext = 0;
    k++;

    for(i=0;i<nbfront;++i) {
      x = FRONT[i].x;
      y = FRONT[i].y;
      z = FRONT[i].z;
      
      px = pere[x][y][z]->x;
      py = pere[x][y][z]->y;
      pz = pere[x][y][z]->z;

      if(V[x+1][y][z] == UNVISITED ) {
	V[x+1][y][z] = V[x][y][z] + 1;
	NEXT[nbnext].x = x+1;
	NEXT[nbnext].y = y;
	NEXT[nbnext].z = z;
	nbnext++;
	pere[x+1][y][z] = pere[x][y][z];
      } else 
	if( abs(pere[x+1][y][z]->x - px) + abs(pere[x+1][y][z]->y - py) + 
	   abs(pere[x+1][y][z]->z - pz)> 3) 
	  voro[x+1][y][z] = 1; 
      
      if(V[x-1][y][z] == UNVISITED )  {
	V[x-1][y][z] = V[x][y][z] + 1;
	NEXT[nbnext].x = x-1;
	NEXT[nbnext].y = y;
	NEXT[nbnext].z = z;
	nbnext++;
	pere[x-1][y][z] = pere[x][y][z];
      } else if( abs(pere[x-1][y][z]->x - px) + abs(pere[x-1][y][z]->y - py) 
		+ abs(pere[x-1][y][z]->z - pz)> 3) 
	voro[x-1][y][z] = 1; 

      if(V[x][y+1][z] == UNVISITED )  {
	V[x][y+1][z] = V[x][y][z] + 1;
	NEXT[nbnext].x = x;
	NEXT[nbnext].y = y+1;
	NEXT[nbnext].z = z;
	nbnext++;
	pere[x][y+1][z] = pere[x][y][z];
      }else if( abs(pere[x][y+1][z]->x - px) + abs(pere[x][y+1][z]->y - py) 
	       + abs(pere[x][y+1][z]->z - pz)> 3) 
	voro[x][y+1][z] = 1; 
      
      if(V[x][y-1][z] == UNVISITED )  {
	V[x][y-1][z] = V[x][y][z] + 1;
	NEXT[nbnext].x = x;
	NEXT[nbnext].y = y-1;
	NEXT[nbnext].z = z;
	nbnext++;
	pere[x][y-1][z] = pere[x][y][z];
      } else if( abs(pere[x][y-1][z]->x - px) + abs(pere[x][y-1][z]->y - py) 
		+ abs(pere[x][y-1][z]->z - pz)> 3) 
	voro[x][y-1][z] = 1; 

      if(V[x][y][z+1] == UNVISITED )  {
	V[x][y][z+1] = V[x][y][z] + 1;
	NEXT[nbnext].x = x;
	NEXT[nbnext].y = y;
	NEXT[nbnext].z = z+1;
	nbnext++;
	pere[x][y][z+1] = pere[x][y][z];
      } else if( abs(pere[x][y][z+1]->x - px) + abs(pere[x][y][z+1]->y - py) 
		+ abs(pere[x][y][z+1]->z - pz)> 3) 
	voro[x][y][z+1] = 1; 

      if(V[x][y][z-1] == UNVISITED )  {
	V[x][y][z-1] = V[x][y][z] + 1;
	NEXT[nbnext].x = x;
	NEXT[nbnext].y = y;
	NEXT[nbnext].z = z-1;
	nbnext++;
	pere[x][y][z-1] = pere[x][y][z];
      } else if( abs(pere[x][y][z-1]->x - px) + abs(pere[x][y][z-1]->y - py) 
		+ abs(pere[x][y][z-1]->z - pz)> 3) 
	voro[x][y][z-1] = 1; 
    }
    TEMP = FRONT;
    FRONT = NEXT;
    NEXT = TEMP;
    nbfront = nbnext;

#ifdef VERBOSE
    if (verbose)
      fprintf(stderr,"wavefront #%d:\tnbfront:%d.\n",k,nbfront);
#endif
#ifdef DEBUG
      fprintf(log_fd,"wavefront #%d:\tnbfront:%d.\n",k,nbfront);
#endif

  }

  /* free the memory */

  free(FRONT);
  free(NEXT);
  free(pere[0][0]);
  free(pere[0]);
  free(pere);
  free(ppepere);

}/*END*/



