/*****************************************************************************

  3D Randomized Motion Planner

  COPYRIGHT (c) 1990 
  THE BOARD OF TRUSTEES OF THE LELAND STANFORD JUNIOR UNIVERSITY
  ALL RIGHTS RESERVED

  No duplications, whole or partial, manual or electronic, may be made 
  without the express permission of the author. Any such copies, or 
  revisions thereof, must display this notice unaltered.

  Robotics Laboratory,
  Computer Science Department,
  Stanford University

  Original author: Jerome Barraquand
  Modified by Tsai-Yen Li (li@flamingo.Stanford.EDU)
  Advisor: Jean-Claude Latombe

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "defs.h"
#include "solidstruct.h"
#include "macro.h"
#include "solidp.h"

/****************************************************************************
  declare local constants
****************************************************************************/

#define MAX_FRAME 30

/****************************************************************************
  Static global variables
****************************************************************************/

static fpoint3D Iunit = {1.0, 0.0, 0.0} ;
static fpoint3D Junit = {0.0, 1.0, 0.0} ;
static fpoint3D Kunit = {0.0, 0.0, 1.0} ;
static float *FK_cos,*FK_sin;

#ifndef EXACT
static fpoint3D ***FKI;
static fpoint3D ***FKJ;
static fpoint3D ***FKK;
#endif

/****************************************************************************
  external variables
****************************************************************************/

extern robot therobot;
extern descript ws_des;
extern FILE *fd;
extern long nb_fk_calls;

/****************************************************************************
  discrete version of rotation transformation                     
****************************************************************************/

/****************************************************************************
  Rotate a point along x-axis by theta units 
  Given P, rotate theta, return Q.
****************************************************************************/

void RX(int theta,fpoint3D *P,fpoint3D *Q) {

  Q->x = P->x; 
  Q->y = FK_cos[theta] * P->y - FK_sin[theta] * P->z;
  Q->z = FK_sin[theta] * P->y + FK_cos[theta] * P->z;
}


/****************************************************************************
  Rotate a point along y-axis by theta units 
  Given P, rotate theta, return Q.
****************************************************************************/

void RY(int theta,fpoint3D *P,fpoint3D *Q){

  Q->y = P->y; 
  Q->z = FK_cos[theta] * P->z - FK_sin[theta] * P->x;
  Q->x = FK_sin[theta] * P->z + FK_cos[theta] * P->x;
}

/****************************************************************************
 Rotate a point along z-axis by theta units 
  Given P, rotate theta, return Q.
****************************************************************************/

void RZ(int theta,fpoint3D *P,fpoint3D *Q){
  
  Q->z = P->z; 
  Q->x = FK_cos[theta] * P->x - FK_sin[theta] * P->y;
  Q->y = FK_sin[theta] * P->x + FK_cos[theta] * P->y;
}


/****************************************************************************
  Initialize the forward kinematics lookup table                   
  Initialize the frame unit vector of 3d array
  The convention is raw, pitch, yowl.
  The data will be used in path search: down or brownian motion.
****************************************************************************/

void Init_FK(void) {

  register int i,dimc=therobot.dimc[0];

  /* allocate memory for sin and cos lookup table */
  FK_cos = (float *)calloc(therobot.dimc[0],sizeof(float));
  FK_sin = (float *)calloc(therobot.dimc[0],sizeof(float));

  /* build the lookup table for sin and cos first */
  for(i=0;i<dimc;i++) {
    FK_cos[i]=(float)cos(therobot.coefc[0]*i*TWOPI);
    FK_sin[i]=(float)sin(therobot.coefc[0]*i*TWOPI);
  }

#ifndef EXACT
  /* allocate memory for fk lookup table */
  FKI = (fpoint3D ***) callocf3D(dimc,dimc,dimc,1);
  FKJ = (fpoint3D ***) callocf3D(dimc,dimc,dimc,1);
  FKK = (fpoint3D ***) callocf3D(dimc,dimc,dimc,1);

  /* Initialize the lookup table */
  for (i=0;i<dimc;i++) {
    RX(i,&FK_Iunit,&FK_Itemp1);  
    RX(i,&FK_Junit,&FK_Jtemp1); 
    RX(i,&FK_Kunit,&FK_Ktemp1);  
    for (j=0;j<therobot.dimc[0];j++) {
      RY(j,&FK_Itemp1,&FK_Itemp2);  
      RY(j,&FK_Jtemp1,&FK_Jtemp2);  
      RY(j,&FK_Ktemp1,&FK_Ktemp2);  
      for (k=0;k<therobot.dimc[0];k++) {
	RZ(k,&FK_Itemp2,&FKI[i][j][k]);  
	RZ(k,&FK_Jtemp2,&FKJ[i][j][k]); 
	RZ(k,&FK_Ktemp2,&FKK[i][j][k]);  
      }
    }
  }
#endif

}

/****************************************************************************
  Continuous version of rotating a vector along axises
  To gain some speedup, there can be replaced by macros.
****************************************************************************/

/****************************************************************************
  rotate a point P along x-axis by theta radian
  Given P, rotate theta, return Q.
****************************************************************************/

void CRX(double theta,fpoint3D *P,fpoint3D *Q) {

  double sin_theta=sin(theta),cos_theta=cos(theta);

  Q->x = P->x; 
  Q->y = cos_theta * P->y - sin_theta * P->z;
  Q->z = sin_theta * P->y + cos_theta * P->z;
}


/****************************************************************************
  rotate a point P along y-axis by theta radian 
  Given P, rotate theta, return Q.
****************************************************************************/

void CRY(double theta,fpoint3D *P,fpoint3D *Q){

  double sin_theta=sin(theta),cos_theta=cos(theta);

  Q->y = P->y; 
  Q->z = cos_theta * P->z - sin_theta * P->x;
  Q->x = sin_theta * P->z + cos_theta * P->x;
}

/****************************************************************************
  rotate a point P along z-axis by theta radian 
  Given P, rotate theta, return Q.
****************************************************************************/

void CRZ(double theta,fpoint3D *P,fpoint3D *Q){
  
  double sin_theta=sin(theta),cos_theta=cos(theta);

  Q->z = P->z; 
  Q->x = cos_theta * P->x - sin_theta * P->y;
  Q->y = sin_theta * P->x + cos_theta * P->y;
}

/****************************************************************************
  find the unit vectors of new frame after certain rotations
  used in bitmap construction (synth.c)
****************************************************************************/

void FK_Unit(configuration X,fpoint3D Y[3]) {

  double theta,phi,psi,deg_rad=M_PI/180.0;
  fpoint3D temp,temp2;

  theta = deg_rad*X[3];
  phi   = deg_rad*X[4];
  psi   = deg_rad*X[5]; 
 
  CRX( theta, &Iunit, &temp); 
  CRY( phi, &temp, &temp2);  
  CRZ( psi, &temp2, &(Y[0]));  
 
  CRX( theta, &Junit, &temp);  
  CRY( phi, &temp, &temp2);  
  CRZ( psi, &temp2, &(Y[1]));  
 
  CRX( theta, &Kunit, &temp); 
  CRY( phi, &temp, &temp2); 
  CRZ( psi, &temp2, &(Y[2])); 

  return;
}

/****************************************************************************
  Find the foreward kinematics points
  Given OldP, and a configuration (6D), return NewP
****************************************************************************/

void FK_Point(configuration T,fpoint3D OldP,fpoint3D *NewP) {

  double theta,phi,psi,deg_rad=M_PI/180.0;
  fpoint3D temp,temp2;

  theta = deg_rad*T[3];
  phi   = deg_rad*T[4];
  psi   = deg_rad*T[5]; 
 
  CRX( theta, &OldP, &temp); 
  CRY( phi, &temp, &temp2);  
  CRZ( psi, &temp2, NewP);  
 
  NewP->x += T[0];
  NewP->y += T[1];
  NewP->z += T[2];
  
  return;
}

/****************************************************************************
  Find the backward kinematics points
  Given OldP, and a configuration (6D), return NewP
****************************************************************************/

void BK_Point(configuration T,fpoint3D OldP,fpoint3D *NewP) {

  double theta,phi,psi,deg_rad=M_PI/180.0;
  fpoint3D temp,temp2;

  theta = deg_rad*T[3];
  phi   = deg_rad*T[4];
  psi   = deg_rad*T[5]; 
 
  OldP.x -= T[0];
  OldP.y -= T[1];
  OldP.z -= T[2];
  
  CRZ( -psi, &OldP, &temp);  
  CRY( -phi, &temp, &temp2);  
  CRX( -theta, &temp2, NewP); 
 
  return;
}

/****************************************************************************
  multiply two matrices 
  C=A*B
  To gain some speedup, may be replaced by macros.
****************************************************************************/

void T_multiply(fvector3D A[3],fvector3D B[3],fvector3D C[3]) {
  
  C[0].x = A[0].x*B[0].x+A[1].x*B[0].y+A[2].x*B[0].z;
  C[0].y = A[0].y*B[0].x+A[1].y*B[0].y+A[2].y*B[0].z;
  C[0].z = A[0].z*B[0].x+A[1].z*B[0].y+A[2].z*B[0].z;
  
  C[1].x = A[0].x*B[1].x+A[1].x*B[1].y+A[2].x*B[1].z;
  C[1].y = A[0].y*B[1].x+A[1].y*B[1].y+A[2].y*B[1].z;
  C[1].z = A[0].z*B[1].x+A[1].z*B[1].y+A[2].z*B[1].z;
  
  C[2].x = A[0].x*B[2].x+A[1].x*B[2].y+A[2].x*B[2].z;
  C[2].y = A[0].y*B[2].x+A[1].y*B[2].y+A[2].y*B[2].z;
  C[2].z = A[0].z*B[2].x+A[1].z*B[2].y+A[2].z*B[2].z;
}

/****************************************************************************
  Find the workspace integer coordinates of the leading points.
  The global coordinates of leading points is calculated from the 
  reference frame and local coordinates in the frame.
  (Global variable "therobot" used)
****************************************************************************/

void trans_lead(ipoint3D *ilead_pts) {

  fpoint3D dlead_pts[MAXLEAD];
  int i,whichframe;
  
  for(i=0;i<(therobot.nblead);i++) {
    whichframe=therobot.leadpoint[i].refframe;
    Local_to_global(therobot.localframe[whichframe].gorig,
		    therobot.localframe[whichframe].unit,
		    therobot.leadpoint[i].localp,
		    &(dlead_pts[i]));
    ilead_pts[i].x=(int)(dlead_pts[i].x*(double)(ws_des.dimx));
    ilead_pts[i].y=(int)(dlead_pts[i].y*(double)(ws_des.dimy));
    ilead_pts[i].z=(int)(dlead_pts[i].z*(double)(ws_des.dimz));
    }
}

/****************************************************************************
  Convert the local coordinates of a point with respecting to the local frame
  to global coordinate.
  Given the origin, unit vectors of the local frame and local coordinate,
  return the global coordinates.
  To speedup, may be replaced by macros.
****************************************************************************/

void Local_to_global(fpoint3D origin,fvector3D unit[3],fpoint3D localp,
		     fpoint3D *globalp) {

  globalp->x=
    localp.x*unit[0].x+localp.y*unit[1].x+localp.z*unit[2].x+origin.x;
  globalp->y=
    localp.x*unit[0].y+localp.y*unit[1].y+localp.z*unit[2].y+origin.y;
  globalp->z=
    localp.x*unit[0].z+localp.y*unit[1].z+localp.z*unit[2].z+origin.z;
}

/***************************************************************************
  forward kinematics calculation:
  Given: configuration X
  To find: forward kinematics of end points of the linkages
  Here I try to be more general in the number of DOF, but there are some
  overhead associated with it. Speedup can be gained by modifying the code
  to be more example specific.
***************************************************************************/

void FK(configuration X,fpoint3D *curpoints){

  int i,i1,theta,phi,psi,dof=0,whichframe;
  fpoint3D tempframe[3],temp1,temp2;

  nb_fk_calls++;
  for(i=1;i<(therobot.nbframe);i++) {
    
    i1=therobot.localframe[i].refframe;
    if (therobot.localframe[i].dofmask & (1)) {
      therobot.localframe[i].lorig.x=X[dof];
      dof++;
    }
    if (therobot.localframe[i].dofmask & (1<<1)) {
      therobot.localframe[i].lorig.y=X[dof];
      dof++;
    }
    if (therobot.localframe[i].dofmask & (1<<2)) {
      therobot.localframe[i].lorig.z=X[dof];
      dof++;
    }
    Local_to_global(therobot.localframe[i1].gorig,
		    therobot.localframe[i1].unit,
		    therobot.localframe[i].lorig,
		    &(therobot.localframe[i].gorig));
    
    if (therobot.localframe[i].dofmask & (1<<3)) {
      theta = (int) (therobot.dimc[dof]*X[dof]);
      dof++;
    } else
      theta = 0;
    
    if (therobot.localframe[i].dofmask & (1<<4)) {
      phi = (int) (therobot.dimc[dof]*X[dof]);
      dof++;
    } else
      phi = 0;
    
    if (therobot.localframe[i].dofmask & (1<<5)) {
      psi = (int) (therobot.dimc[dof]*X[dof]);
      dof++;
    } else
      psi = 0;
    
#ifdef EXACT
    RX(theta,&Iunit,&temp1);  
    RY(phi,&temp1,&temp2); 
    RZ(psi,&temp2,&(tempframe[0]));  
    
    RX(theta,&Junit,&temp1);  
    RY(phi,&temp1,&temp2);  
    RZ(psi,&temp2,&(tempframe[1]));  
  
    RX(theta,&Kunit,&temp1);  
    RY(phi,&temp1,&temp2); 
    RZ(psi,&temp2,&(tempframe[2]));  
#else
    assign_pos(FKI[theta][phi][psi],tempframe[0]);
    assign_pos(FKJ[theta][phi][psi],tempframe[1]);
    assign_pos(FKK[theta][phi][psi],tempframe[2]);
#endif
    T_multiply(therobot.localframe[i1].unit,tempframe,
	       therobot.localframe[i].unit);
  }

  for(i=0;i<(therobot.nbpoint);i++) {
    whichframe=therobot.robotpoint[i].refframe;
    Local_to_global(therobot.localframe[whichframe].gorig,
		    therobot.localframe[whichframe].unit,
		    therobot.robotpoint[i].localp,
		    &(curpoints[i]));
  }
}

/****************************************************************************
  Continuous version of forward kinematics 
  Needed for smoothing.
****************************************************************************/

void CFK(configuration X,fpoint3D *curpoints){

  int i,i1,dof=0,whichframe;
  double theta,phi,psi;
  fvector3D tempframe[3],temp1,temp2;

  for(i=1;i<therobot.nbframe;i++) {
    
    i1=therobot.localframe[i].refframe;
    if (therobot.localframe[i].dofmask & (1)) {
      therobot.localframe[i].lorig.x=X[dof];
      dof++;
    }
    if (therobot.localframe[i].dofmask & (1<<1)) {
      therobot.localframe[i].lorig.y=X[dof];
      dof++;
    }
    if (therobot.localframe[i].dofmask & (1<<2)) {
      therobot.localframe[i].lorig.z=X[dof];
      dof++;
    }

    Local_to_global(therobot.localframe[i1].gorig,
		    therobot.localframe[i1].unit,
		    therobot.localframe[i].lorig,
		    &(therobot.localframe[i].gorig));
    
    if (therobot.localframe[i].dofmask & (1<<3)) {
      theta = (double) (X[dof]*TWOPI);
      dof++;
    } else
      theta = 0.0;
    
    if (therobot.localframe[i].dofmask & (1<<4)) {
      phi = (double) (X[dof]*TWOPI);
      dof++;
    } else
      phi = 0.0;
    
    if (therobot.localframe[i].dofmask & (0x01<<5)) {
      psi = (double) (X[dof]*TWOPI);
      dof++;
    } else
      psi = 0.0;
    
    CRX(theta,&Iunit,&temp1);  
    CRY(phi,&temp1,&temp2); 
    CRZ(psi,&temp2,&(tempframe[0]));  
    
    CRX(theta,&Junit,&temp1);  
    CRY(phi,&temp1,&temp2);  
    CRZ(psi,&temp2,&(tempframe[1]));  
  
    CRX(theta,&Kunit,&temp1);  
    CRY(phi,&temp1,&temp2); 
    CRZ(psi,&temp2,&(tempframe[2]));  
  
    T_multiply(therobot.localframe[i1].unit,tempframe,
	       therobot.localframe[i].unit);
  }

  for(i=0;i<(therobot.nbpoint);i++) {
    whichframe=therobot.robotpoint[i].refframe;
    Local_to_global(therobot.localframe[whichframe].gorig,
		    therobot.localframe[whichframe].unit,
		    therobot.robotpoint[i].localp,
		    &(curpoints[i]));
  }

}
