/*****************************************************************************

  3D Randomized Motion Planner

  COPYRIGHT (c) 1990 
  THE BOARD OF TRUSTEES OF THE LELAND STANFORD JUNIOR UNIVERSITY
  ALL RIGHTS RESERVED

  No duplications, whole or partial, manual or electronic, may be made 
  without the express permission of the author. Any such copies, or 
  revisions thereof, must display this notice unaltered.

  Robotics Laboratory,
  Computer Science Department,
  Stanford University

  Original author: Jerome Barraqua
  Modified by Tsai-Yen Li (li@flamingo.Stanford.EDU)
  Advisor: Jean-Claude Latombe

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "defs.h"
#include "solidstruct.h"
#include "macro.h"
#include "solidp.h"

/* radius around the goals to compute potentials other than the first one */

#define POT_RAD_FACTOR        1.5

/* external global variables */

extern robot therobot;
extern descript ws_des;
extern FILE *log_fd;

/*****************************************************************************
  function to compute the heuristic potential
******************************************************************************/

void heuristic(unsigned char ***voro,unsigned char ***V,ipoint3D *goal,
	       image *iheurpot) {

  register int i,j;
  register int x,y,z;
  int xnew,ynew,znew;
  pixel *FRONT,*NEXT,*SAVOFRONT,*SAVONEXT,*TEMP;
  int nbfront = 0, nbnext = 0;
  int nbsavofront, nbsavonext;
  unsigned char ****heurpot,***vorocopy;
  int heur_value;
  int k,l,m;
  int Vmax;
  ipoint3D goalshift[MAXLEAD],shiftmax[MAXLEAD],shiftmin[MAXLEAD];
  int radius[MAXLEAD-1];
  float factor;
  clock_t start, finish;

  start=Start_Clock("heuristic");
  
  /* allocate memory for heuristic potential field and vorocopy */

  heurpot = (unsigned char ****)calloc(therobot.nblead,sizeof(unsigned char *));

  for (i=0;i<therobot.nblead;i++) {
    if ((heurpot[i] = (unsigned char ***)calloc3D(ws_des.dimx,ws_des.dimy,
						  ws_des.dimz,
						  sizeof(unsigned char))) 
	== NULL){
      perror("Memory allocation error");
      exit(-1);
    }
  }
  
  if ((vorocopy = (unsigned char ***)calloc3D(ws_des.dimx,ws_des.dimy,
					      ws_des.dimz,
					      sizeof(unsigned char))) == NULL){
    perror("Memory allocation error");
    exit(-1);
  }

  /* The first potential field is build on the workspace bitmap */
  /* To set this up, UNVISITED and FREESPACE should have the same value */
  memcpy(&(heurpot[0][0][0][0]),&(V[0][0][0]),
	 ws_des.dimx*ws_des.dimy*ws_des.dimz);
  /* vorocopy is to store the skeleton after connecting to the goal */
  memcpy(&(vorocopy[0][0][0]),&(voro[0][0][0]),
	 ws_des.dimx*ws_des.dimy*ws_des.dimz);
  
  for(i=1;i<ws_des.dimx-1;i++) 
    for(j=1;j<ws_des.dimy-1;j++) 
      for(k=1;k<ws_des.dimz-1;k++) 
	if (heurpot[0][i][j][k]!=OBSTACLE) heurpot[0][i][j][k]=UNVISITED;
  
  /* allocate memory for FRONT and NEXT and the saved version of them */
  if((FRONT=(pixel *)calloc(10*ws_des.dimy*ws_des.dimz,sizeof(pixel)))== NULL){
    perror("Memory allocation error");
    exit(-1);
  }
  
  if((NEXT=(pixel *)calloc(10*ws_des.dimy*ws_des.dimz,sizeof(pixel)))== NULL){
    perror("Memory allocation error");
    exit(-1);
  }
  
  if((SAVOFRONT=(pixel *)calloc(20*ws_des.dimy*ws_des.dimz,sizeof(pixel)))
     ==NULL) {
    perror("Memory allocation error");
    exit(-1);
  }

  if((SAVONEXT=(pixel *)calloc(20*ws_des.dimy*ws_des.dimz,sizeof(pixel)))
     ==NULL) {
    perror("Memory allocation error");
    exit(-1);
  }

  /* initialize variables */

  heurpot[0][goal[0].x][goal[0].y][goal[0].z] = 0;
  x = goal[0].x;
  y = goal[0].y;
  z = goal[0].z;

#ifdef VERBOSE
  if (verbose)
    fprintf(stderr,"goal 0: (%d, %d, %d)\n",x,y,z);
#endif
#ifdef DEBUG
    fprintf(log_fd,"goal 0: (%d, %d, %d)\n",x,y,z);
#endif

  /* connect the goal to the voronoi diagram by following the steepest
   gradient */

#ifdef VERBOSE
  if (verbose)
    fprintf(stderr,"Connecting the goal to the skeleton\n");
#endif

  while(vorocopy[x][y][z] == 0){
    Vmax = -100000;
    for(k= -1;k<=1;++k){
      for(l= -1;l<=1;++l){
	for(m= -1;m<=1;++m){
	  if(k==0 || l==0 || m==0) continue;
	  if(Vmax < (int)V[x+k][y+l][z+m]) { 
	    Vmax = (int)V[x+k][y+l][z+m]; 
	    xnew = x+k; 
	    ynew = y+l; 
	    znew = z+m;
	  }
	}
      }
    }
    vorocopy[x][y][z] = 1;
    x = xnew; y = ynew; z = znew;
  }
  
  /* initialize FRONT list at goal */

  FRONT[0].x = goal[0].x;
  FRONT[0].y = goal[0].y;
  FRONT[0].z = goal[0].z;
  nbfront = 1;
  nbsavofront = 0;

  /* Then follows the expansion on the voronoi diagramm */
#ifdef VERBOSE
  if (verbose)
    fprintf(stderr,"Expanding the potential along the skeleton.\n");
#endif

  while(nbfront > 0) {
    
    nbnext = 0;
    
    for(i=0;i<nbfront;++i){
      
      x = FRONT[i].x;
      y = FRONT[i].y;
      z = FRONT[i].z;
      
      SAVOFRONT[nbsavofront].x = x;
      SAVOFRONT[nbsavofront].y = y;
      SAVOFRONT[nbsavofront].z = z;
      nbsavofront++;
      
      heur_value=(heurpot[0][x][y][z] < OBSTACLE? 
		  (heurpot[0][x][y][z] + 1) : OBSTACLE);
      
      for(k = -1;k<= 1;++k){
	for(l= -1;l<= 1;++l){
	  for(m= -1;m<= 1;++m){
	    if(vorocopy[x+k][y+l][z+m] == 0) continue;
	    if(heurpot[0][x+k][y+l][z+m] == UNVISITED){
	      heurpot[0][x+k][y+l][z+m] = heur_value;
	      NEXT[nbnext].x = x+k;
	      NEXT[nbnext].y = y+l;
	      NEXT[nbnext].z = z+m;
	      nbnext++;
	    }
	  }
	}
      }
    }
    
    TEMP = FRONT;
    FRONT = NEXT;
    NEXT = TEMP;
    nbfront = nbnext;
  }

  free(FRONT);
  free(NEXT);
  free(vorocopy[0][0]);
  free(vorocopy[0]);
  free(vorocopy);
  
  /* Then follows the expansion from the voronoi diagram to the whole 
     remaining space */

#ifdef VERBOSE
  if (verbose)
    fprintf(stderr,"Expanding the potential to the rest of the workspace.\n");
#endif

  while(nbsavofront){
    
    nbsavonext = 0;
    
    for(i=0;i<nbsavofront;++i){
      x = SAVOFRONT[i].x;
      y = SAVOFRONT[i].y;
      z = SAVOFRONT[i].z;
            
      heur_value=(heurpot[0][x][y][z] < OBSTACLE? 
		  (heurpot[0][x][y][z] + 1) : OBSTACLE);
      
      if(heurpot[0][x+1][y][z] == UNVISITED){
	heurpot[0][x+1][y][z] = heur_value;
	SAVONEXT[nbsavonext].x = x+1;
	SAVONEXT[nbsavonext].y = y;
	SAVONEXT[nbsavonext].z = z;
	nbsavonext++;
      }
      
      if(heurpot[0][x-1][y][z] == UNVISITED){
	heurpot[0][x-1][y][z] = heur_value;
	SAVONEXT[nbsavonext].x = x-1;
	SAVONEXT[nbsavonext].y = y;
	SAVONEXT[nbsavonext].z = z;
	nbsavonext++;
      }
      
      if(heurpot[0][x][y+1][z] == UNVISITED){
	heurpot[0][x][y+1][z] = heur_value;
	SAVONEXT[nbsavonext].x = x;
	SAVONEXT[nbsavonext].y = y+1;
	SAVONEXT[nbsavonext].z = z;
	nbsavonext++;
      }
      
      if(heurpot[0][x][y-1][z] == UNVISITED){
	heurpot[0][x][y-1][z] = heur_value;
	SAVONEXT[nbsavonext].x = x;
	SAVONEXT[nbsavonext].y = y-1;
	SAVONEXT[nbsavonext].z = z;
	nbsavonext++;
      }
      
      if(heurpot[0][x][y][z+1] == UNVISITED){
	heurpot[0][x][y][z+1] = heur_value;
	SAVONEXT[nbsavonext].x = x;
	SAVONEXT[nbsavonext].y = y;
	SAVONEXT[nbsavonext].z = z+1;
	nbsavonext++;
      }
      
      if(heurpot[0][x][y][z-1] == UNVISITED){
	heurpot[0][x][y][z-1] = heur_value;
	SAVONEXT[nbsavonext].x = x;
	SAVONEXT[nbsavonext].y = y;
	SAVONEXT[nbsavonext].z = z-1;
	nbsavonext++;
      }
    }
    TEMP = SAVOFRONT;
    SAVOFRONT = SAVONEXT;
    SAVONEXT = TEMP;
    nbsavofront = nbsavonext;
  }

  /* The end of the potential construction for the first control point */

  /* The potetials for other control points are copied from the first one 
     and only recomputed around the goal */

#ifdef VERBOSE
  if (verbose)
    fprintf(stderr,"goal 1: (%d, %d, %d)\n",goal[1].x,goal[1].y,goal[1].z);
#endif
#ifdef DEBUG
    fprintf(log_fd,"goal 1: (%d, %d, %d)\n",goal[1].x,goal[1].y,goal[1].z);
#endif

  /* Compute the other potential fields by copying the first one onto
     the others and modify the potential within some distance around 
     the goal. This is to save some time but may create extra local minima */

  for (i=1;i<therobot.nblead;i++) 
    memcpy(&(heurpot[i][0][0][0]),&(heurpot[0][0][0][0]),
	   ws_des.dimx*ws_des.dimy*ws_des.dimz);

  factor=POT_RAD_FACTOR;
  for(i=1;i<therobot.nblead;i++) {
    goalshift[i-1].x=abs(goal[0].x-goal[i].x);
    goalshift[i-1].y=abs(goal[0].y-goal[i].y);
    goalshift[i-1].z=abs(goal[0].z-goal[i].z);
    radius[i-1]=(int)sqrt((double)(goalshift[i-1].x*goalshift[i-1].x+
		   goalshift[i-1].y*goalshift[i-1].y+
		   goalshift[i-1].z*goalshift[i-1].z));

#ifdef VERBOSE
    if (verbose) 
      fprintf(stderr,"Rebuilding the potential around the goal.\n");
#endif

#ifdef DEBUG
      fprintf(log_fd,"radius of rebuild = %d\n",radius[i-1]);
#endif

    shiftmin[i-1].x=MAX(goal[i].x-(int)(factor * radius[i-1]),1);
    shiftmin[i-1].y=MAX(goal[i].y-(int)(factor * radius[i-1]),1);
    shiftmin[i-1].z=MAX(goal[i].z-(int)(factor * radius[i-1]),1);
    shiftmax[i-1].x=MIN(goal[i].x+(int)(factor * radius[i-1]),ws_des.dimx-2);
    shiftmax[i-1].y=MIN(goal[i].y+(int)(factor * radius[i-1]),ws_des.dimy-2);
    shiftmax[i-1].z=MIN(goal[i].z+(int)(factor * radius[i-1]),ws_des.dimz-2);

#ifdef VERBOSE
    if (verbose) {
      fprintf(stderr,"potential #%d rebuilding block:\n",i);
      fprintf(stderr,"xmin=%d,xmax=%d\nymin=%d,ymax=%d\nzmin=%d,zmax=%d\n",
	      shiftmin[i-1].x,shiftmax[i-1].x,shiftmin[i-1].y,shiftmax[i-1].y,
	      shiftmin[i-1].z,shiftmax[i-1].z);
    }
#endif
#ifdef DEBUG
      fprintf(log_fd,"potential #%d rebuilding block:\n",i);
      fprintf(log_fd,"xmin=%d,xmax=%d\nymin=%d,ymax=%d\nzmin=%d,zmax=%d\n",
	      shiftmin[i-1].x,shiftmax[i-1].x,shiftmin[i-1].y,shiftmax[i-1].y,
	      shiftmin[i-1].z,shiftmax[i-1].z);
#endif

    /* mark it unvisited first */
    for(x=shiftmin[i-1].x;x<=shiftmax[i-1].x;x++) {
      for(y=shiftmin[i-1].y;y<=shiftmax[i-1].y;y++) {
	for(z=shiftmin[i-1].z;z<=shiftmax[i-1].z;z++) {
	  if(V[x][y][z]!=OBSTACLE) heurpot[i][x][y][z]=UNVISITED;
	}
      }
    }
    
    /* rebuild the potential by expanding wavefronts from the goal */
    heurpot[i][goal[i].x][goal[i].y][goal[i].z]=0;
    SAVOFRONT[0].x=goal[i].x;
    SAVOFRONT[0].y=goal[i].y;
    SAVOFRONT[0].z=goal[i].z;
    nbsavofront=1;
    while(nbsavofront) {
      nbsavonext = 0;
    
      for(j=0;j<nbsavofront;++j){
	x = SAVOFRONT[j].x;
	y = SAVOFRONT[j].y;
	z = SAVOFRONT[j].z;

	heur_value = (heurpot[i][x][y][z] < OBSTACLE?heurpot[i][x][y][z] + 1:
		 OBSTACLE);
      
	if(heurpot[i][x+1][y][z] == UNVISITED){
	  heurpot[i][x+1][y][z] = heur_value;
	  SAVONEXT[nbsavonext].x = x+1;
	  SAVONEXT[nbsavonext].y = y;
	  SAVONEXT[nbsavonext].z = z;
	  nbsavonext++;
	}
	
	if(heurpot[i][x-1][y][z] == UNVISITED){
	  heurpot[i][x-1][y][z] = heur_value;
	  SAVONEXT[nbsavonext].x = x-1;
	  SAVONEXT[nbsavonext].y = y;
	  SAVONEXT[nbsavonext].z = z;
	  nbsavonext++;
	}
	
	if(heurpot[i][x][y+1][z] == UNVISITED){
	  heurpot[i][x][y+1][z] = heur_value;
	  SAVONEXT[nbsavonext].x = x;
	  SAVONEXT[nbsavonext].y = y+1;
	  SAVONEXT[nbsavonext].z = z;
	  nbsavonext++;
	}
	
	if(heurpot[i][x][y-1][z] == UNVISITED){
	  heurpot[i][x][y-1][z] = heur_value;
	  SAVONEXT[nbsavonext].x = x;
	  SAVONEXT[nbsavonext].y = y-1;
	  SAVONEXT[nbsavonext].z = z;
	  nbsavonext++;
	}
	
	if(heurpot[i][x][y][z+1] == UNVISITED){
	  heurpot[i][x][y][z+1] = heur_value;
	  SAVONEXT[nbsavonext].x = x;
	  SAVONEXT[nbsavonext].y = y;
	  SAVONEXT[nbsavonext].z = z+1;
	  nbsavonext++;
	}
	
	if(heurpot[i][x][y][z-1] == UNVISITED){
	  heurpot[i][x][y][z-1] = heur_value;
	  SAVONEXT[nbsavonext].x = x;
	  SAVONEXT[nbsavonext].y = y;
	  SAVONEXT[nbsavonext].z = z-1;
	  nbsavonext++;
	}
      }
      TEMP = SAVOFRONT;
      SAVOFRONT = SAVONEXT;
      SAVONEXT = TEMP;
      nbsavofront = nbsavonext;
    }
  }

  /* pass back the potentials */
  for (i=0;i<therobot.nblead;i++) 
    iheurpot[i].d = heurpot[i];

  free(SAVOFRONT);
  free(SAVONEXT);
  finish=Finish_Clock("heuristic");
  Calculate_Time("heuristic",start,finish);

}/* END */

