/*****************************************************************************

  3D Randomized Motion Planner

  COPYRIGHT (c) 1990 
  THE BOARD OF TRUSTEES OF THE LELAND STANFORD JUNIOR UNIVERSITY
  ALL RIGHTS RESERVED

  No duplications, whole or partial, manual or electronic, may be made 
  without the express permission of the author. Any such copies, or 
  revisions thereof, must display this notice unaltered.

  Robotics Laboratory,
  Computer Science Department,
  Stanford University

  Original author: Jerome Barraquand
  Modified by Tsai-Yen Li (li@flamingo.Stanford.EDU)
  Advisor: Jean-Claude Latombe

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <malloc.h>
#include "defs.h"
#include "solidstruct.h"
#include "macro.h"
#include "solidp.h"
#include "parse.h"

extern robot therobot;
extern descript ws_des;
extern FILE *log_fd;

/*****************************************************************************
  scan until the end of the line 
*****************************************************************************/

void fscanf_eol(FILE *fd)  {
  char ch;
  do {
    ch=fgetc(fd); 
  } while(ch != '\n');

}

/*****************************************************************************
  Read Robot info from file: robot.dat
*****************************************************************************/

void Read_Robot_File(FILE *fd) {

  register int i,j=0,k=0;
  int m[6];

#ifdef DEBUG
  log_fd=fopen("debug.log","w");
#endif

  /* Reading degree of freedom */

  fscanf_eol(fd);
  fscanf(fd,"%d\n",&(therobot.dof));

  if (therobot.dof!=DEGREE) {
    perror("Degree of freedom inconsistent");
    exit(-1);
  }

#ifdef DEBUG
  fprintf(log_fd,"%d\n",therobot.dof);
  if (verbose)
    fprintf(stderr,"%d\n",therobot.dof);
#endif

  /* Reading dimension of C-Space for each DOF */
  therobot.dimc = (int *) calloc(sizeof(int),therobot.dof);
  therobot.rot = (int *) calloc(sizeof(int),therobot.dof);
  therobot.coefc = (double *) calloc(sizeof(double),therobot.dof);

  fscanf_eol(fd);
  for (i=0; i< (therobot.dof); i++) {
    fscanf(fd,"%d",&(therobot.dimc[i]));

#ifdef DEBUG
    fprintf(log_fd,"%d ",therobot.dimc[i]);
  if (verbose)
    fprintf(stderr,"%d ",therobot.dimc[i]);
#endif

    therobot.coefc[i] = 1.0/(double)(therobot.dimc[i]);
  }

#ifdef DEBUG
  NL(log_fd);
  if (verbose)
  NL(stderr);
#endif

  fscanf_eol(fd);

  /* Reading number of points */
  fscanf_eol(fd);
  fscanf(fd,"%d\n",&(therobot.nbpoint));

#ifdef DEBUG
  fprintf(log_fd,"nbp=%d\n",therobot.nbpoint);
  if (verbose)
  fprintf(stderr,"nbp=%d\n",therobot.nbpoint);
#endif

  fscanf_eol(fd);

  /* Reading number of links */
  fscanf(fd,"%d\n",&(therobot.nblink));

#ifdef DEBUG
  fprintf(log_fd,"nblink=%d\n",therobot.nblink);
  if (verbose)
  fprintf(stderr,"nblink=%d\n",therobot.nblink);
#endif

  /* Reading link info */
  therobot.robotlink = (rlink *) calloc(sizeof(rlink),therobot.nblink);
  fscanf_eol(fd);

  /* Reading link length */
  for (i=0; i< (therobot.nblink); i++) {
    fscanf(fd,"%lf",&(therobot.robotlink[i].dlength));
#ifdef DEBUG
    fprintf(log_fd,"%4.2f ",therobot.robotlink[i].dlength);
  if (verbose)
    fprintf(stderr,"%4.2f ",therobot.robotlink[i].dlength);
#endif
    therobot.robotlink[i].ilength = 
      (int)((therobot.robotlink[i].dlength)*ws_des.dimx);
  }

#ifdef DEBUG
  if (verbose)
    NL(stderr);
  NL(log_fd);
#endif
  
  /* Reading link endpoints */
  fscanf_eol(fd);
  for (i=0; i< (therobot.nblink); i++) {
    fscanf_eol(fd);
    fscanf(fd,"%d %d\n",&(therobot.robotlink[i].endp1),
	   &(therobot.robotlink[i].endp2));
#ifdef DEBUG
    fprintf(log_fd,"link #%d = %d %d\n",i,therobot.robotlink[i].endp1,
	   therobot.robotlink[i].endp2);
  if (verbose)
    fprintf(stderr,"link #%d = %d %d\n",i,therobot.robotlink[i].endp1,
	   therobot.robotlink[i].endp2);
#endif
  }

  /* Reading number of frames */
  fscanf_eol(fd);
  fscanf(fd,"%d\n",&(therobot.nbframe));

#ifdef DEBUG
  fprintf(log_fd,"nbframe=%d\n",therobot.nbframe);
  if (verbose)
  fprintf(stderr,"nbframe=%d\n",therobot.nbframe);
#endif

  /* Reading frame info */
  therobot.localframe = (frame *) calloc(sizeof(frame),therobot.nbframe);

  /* Initialize frame #0 */
  therobot.localframe[0].refframe=0;
  therobot.localframe[0].lorig.x=0.0;
  therobot.localframe[0].lorig.y=0.0;
  therobot.localframe[0].lorig.z=0.0;
  therobot.localframe[0].lorie.x=0.0;
  therobot.localframe[0].lorie.y=0.0;
  therobot.localframe[0].lorie.z=0.0;
  therobot.localframe[0].gorig.x=0.0;
  therobot.localframe[0].gorig.y=0.0;
  therobot.localframe[0].gorig.z=0.0;
  therobot.localframe[0].unit[0].x=1.0;
  therobot.localframe[0].unit[0].y=0.0;
  therobot.localframe[0].unit[0].z=0.0;
  therobot.localframe[0].unit[1].x=0.0;
  therobot.localframe[0].unit[1].y=1.0;
  therobot.localframe[0].unit[1].z=0.0;
  therobot.localframe[0].unit[2].x=0.0;
  therobot.localframe[0].unit[2].y=0.0;
  therobot.localframe[0].unit[2].z=1.0;

  for (i=1; i< (therobot.nbframe); i++) {

    /* Reading the reference frame number */
    fscanf_eol(fd);
    fscanf(fd,"%d\n",&(therobot.localframe[i].refframe));

#ifdef DEBUG
    fprintf(log_fd,"ref=%d\n",therobot.localframe[i].refframe);
  if (verbose)
    fprintf(stderr,"ref=%d\n",therobot.localframe[i].refframe);
#endif

    /* Reading the frame origin relative to the reference one */
    fscanf_eol(fd);
    fscanf(fd,"%f %f %f %f %f %f\n",
	   &(therobot.localframe[i].lorig.x),
	   &(therobot.localframe[i].lorig.y),
	   &(therobot.localframe[i].lorig.z),
	   &(therobot.localframe[i].lorie.x),
	   &(therobot.localframe[i].lorie.y),
	   &(therobot.localframe[i].lorie.z));
#ifdef DEBUG
    fprintf(log_fd,"%f %f %f %f %f %f\n",
	   therobot.localframe[i].lorig.x,
	   therobot.localframe[i].lorig.y,
	   therobot.localframe[i].lorig.z,
	   therobot.localframe[i].lorie.x,
	   therobot.localframe[i].lorie.y,
	   therobot.localframe[i].lorie.z);
  if (verbose)
    fprintf(stderr,"%f %f %f %f %f %f\n",
	   therobot.localframe[i].lorig.x,
	   therobot.localframe[i].lorig.y,
	   therobot.localframe[i].lorig.z,
	   therobot.localframe[i].lorie.x,
	   therobot.localframe[i].lorie.y,
	   therobot.localframe[i].lorie.z);
#endif

    fscanf_eol(fd);
    fscanf(fd,"%f %f %f %f %f %f\n",
	   &(therobot.localframe[i].maxlimit[0]),
	   &(therobot.localframe[i].maxlimit[1]),
	   &(therobot.localframe[i].maxlimit[2]),
	   &(therobot.localframe[i].maxlimit[3]),
	   &(therobot.localframe[i].maxlimit[4]),
	   &(therobot.localframe[i].maxlimit[5]));

#ifdef DEBUG
    fprintf(log_fd,"%f %f %f %f %f %f\n",
	   therobot.localframe[i].maxlimit[0],
	   therobot.localframe[i].maxlimit[1],
	   therobot.localframe[i].maxlimit[2],
	   therobot.localframe[i].maxlimit[3],
	   therobot.localframe[i].maxlimit[4],
	   therobot.localframe[i].maxlimit[5]);
  if (verbose)
    fprintf(stderr,"%f %f %f %f %f %f\n",
	   therobot.localframe[i].maxlimit[0],
	   therobot.localframe[i].maxlimit[1],
	   therobot.localframe[i].maxlimit[2],
	   therobot.localframe[i].maxlimit[3],
	   therobot.localframe[i].maxlimit[4],
	   therobot.localframe[i].maxlimit[5]);
#endif

    fscanf_eol(fd);
    fscanf(fd,"%f %f %f %f %f %f\n",
	   &(therobot.localframe[i].minlimit[0]),
	   &(therobot.localframe[i].minlimit[1]),
	   &(therobot.localframe[i].minlimit[2]),
	   &(therobot.localframe[i].minlimit[3]),
	   &(therobot.localframe[i].minlimit[4]),
	   &(therobot.localframe[i].minlimit[5]));

#ifdef DEBUG
    fprintf(log_fd,"%f %f %f %f %f %f\n",
	   therobot.localframe[i].minlimit[0],
	   therobot.localframe[i].minlimit[1],
	   therobot.localframe[i].minlimit[2],
	   therobot.localframe[i].minlimit[3],
	   therobot.localframe[i].minlimit[4],
	   therobot.localframe[i].minlimit[5]);
  if (verbose)
    fprintf(stderr,"%f %f %f %f %f %f\n",
	   therobot.localframe[i].minlimit[0],
	   therobot.localframe[i].minlimit[1],
	   therobot.localframe[i].minlimit[2],
	   therobot.localframe[i].minlimit[3],
	   therobot.localframe[i].minlimit[4],
	   therobot.localframe[i].minlimit[5]);
#endif


    /* Reading DOF mask for the local frame */
    fscanf_eol(fd);
    fscanf(fd,"%d %d %d %d %d %d\n",m,m+1,m+2,m+3,m+4,m+5);

#ifdef DEBUG
    fprintf(log_fd,"dofmask=%d %d %d %d %d %d\n",m[0],m[1],m[2],m[3],m[4],m[5]);
  if (verbose)
    fprintf(stderr,"dofmask=%d %d %d %d %d %d\n",m[0],m[1],m[2],m[3],m[4],m[5]);
#endif

    therobot.localframe[i].dofmask=m[0]+m[1]*2+m[2]*4+m[3]*8+m[4]*16+m[5]*32;
    for(j=0;j<6;j++) {
      if (m[j]==1) {
	if (j>2) therobot.rot[k]=1;
	else therobot.rot[k]=0;
	k++;
      }
    }
  }

  /* Reading robot points info */
  therobot.robotpoint=(refpoint *) calloc(sizeof(refpoint),therobot.nbpoint);
  for (i=0; i<therobot.nbpoint; i++) {

    /* Reading the reference frame number */
    fscanf_eol(fd);
    fscanf(fd,"%d\n",&(therobot.robotpoint[i].refframe));

#ifdef DEBUG
    fprintf(log_fd,"%d\n",therobot.robotpoint[i].refframe);
  if (verbose)
    fprintf(stderr,"%d\n",therobot.robotpoint[i].refframe);
#endif

    /* Reading the local coordinates of the point */
    fscanf_eol(fd);
    fscanf(fd,"%f %f %f\n",
	   &(therobot.robotpoint[i].localp.x),
	   &(therobot.robotpoint[i].localp.y),
	   &(therobot.robotpoint[i].localp.z));
#ifdef DEBUG
    fprintf(log_fd,"%f %f %f\n",
	   therobot.robotpoint[i].localp.x,
	   therobot.robotpoint[i].localp.y,
	   therobot.robotpoint[i].localp.z);
  if (verbose)
    fprintf(stderr,"%f %f %f\n",
	   therobot.robotpoint[i].localp.x,
	   therobot.robotpoint[i].localp.y,
	   therobot.robotpoint[i].localp.z);
#endif
  }

  /* Reading the number of leading points */
  fscanf_eol(fd);
  fscanf(fd,"%d\n",&(therobot.nblead));

#ifdef DEBUG
  fprintf(log_fd,"nblead=%d\n",therobot.nblead);
  if (verbose)
  fprintf(stderr,"nblead=%d\n",therobot.nblead);
#endif

  /* Reading the leading point info */
  therobot.leadpoint = (refpoint *) calloc(sizeof(refpoint),therobot.nblead);
  for (i=0; i< (therobot.nblead); i++) {
    fscanf_eol(fd);
    fscanf(fd,"%d\n",&(therobot.leadpoint[i].refframe));

#ifdef DEBUG
    fprintf(log_fd,"reff=%d\n",therobot.leadpoint[i].refframe);
  if (verbose)
    fprintf(stderr,"reff=%d\n",therobot.leadpoint[i].refframe);
#endif

    fscanf_eol(fd);
    fscanf(fd,"%f %f %f\n",
	   &(therobot.leadpoint[i].localp.x),
	   &(therobot.leadpoint[i].localp.y),
	   &(therobot.leadpoint[i].localp.z));

#ifdef DEBUG
    fprintf(log_fd,"%f %f %f\n",
	   therobot.leadpoint[i].localp.x,
	   therobot.leadpoint[i].localp.y,
	   therobot.leadpoint[i].localp.z);
  if (verbose)
    fprintf(stderr,"%f %f %f\n",
	   therobot.leadpoint[i].localp.x,
	   therobot.leadpoint[i].localp.y,
	   therobot.leadpoint[i].localp.z);
#endif

  }
  fscanf_eol(fd);
  for (i=0;i<therobot.dof;i++) {
    fscanf(fd,"%f", &(therobot.initq[i]));

#ifdef DEBUG
    fprintf(log_fd,"%f ", therobot.initq[i]);
  if (verbose)
    fprintf(stderr,"%f ", therobot.initq[i]);
#endif

  }
  fscanf_eol(fd);

#ifdef DEBUG
  if (verbose)
  NL(stderr);
  NL(log_fd);
#endif

  therobot.goallead= (fpoint3D *) calloc(sizeof(fpoint3D),therobot.nblead);
  for (i=0;i<therobot.nblead;i++) {
    fscanf_eol(fd);
    fscanf(fd,"%f %f %f\n", 
	   &(therobot.goallead[i].x),
	   &(therobot.goallead[i].y),
	   &(therobot.goallead[i].z));

#ifdef DEBUG
    fprintf(log_fd,"%f %f %f\n", 
	   therobot.goallead[i].x,
	   therobot.goallead[i].y,
	   therobot.goallead[i].z);
  if (verbose)
    fprintf(stderr,"%f %f %f\n", 
	   therobot.goallead[i].x,
	   therobot.goallead[i].y,
	   therobot.goallead[i].z);
#endif

  }

#ifdef DEBUG
  if (verbose)
    NL(stderr);
  NL(log_fd);
#endif

}
  




