/*****************************************************************************

  3D Randomized Motion Planner

  COPYRIGHT (c) 1990 
  THE BOARD OF TRUSTEES OF THE LELAND STANFORD JUNIOR UNIVERSITY
  ALL RIGHTS RESERVED

  No duplications, whole or partial, manual or electronic, may be made 
  without the express permission of the author. Any such copies, or 
  revisions thereof, must display this notice unaltered.

  Robotics Laboratory,
  Computer Science Department,
  Stanford University

  Original author: Jerome Barraquand
  Modified by Tsai-Yen Li (li@flamingo.Stanford.EDU)
  Advisor: Jean-Claude Latombe (latombe@flamingo.Stanford.EDU)

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "defs.h"
#include "solidstruct.h"
#include "macro.h"
#include "solidp.h"

/*****************************************************************************
  declare local constants
*****************************************************************************/

#define NBTMAX       2000

/*****************************************************************************
  external variables
*****************************************************************************/

extern unsigned short xsubi[3];
extern robot therobot;
extern descript ws_des;
extern FILE *log_fd;
extern long nb_no_hit;
extern void bcopy();

/*****************************************************************************
  function to find down motion
*****************************************************************************/

boolean down(configuration curq,float *Uend,configuration *result,
	     int *length,image V,image *heurpot) {

  int i,nbtrials;
  boolean found,goodmove,out;
  float Mcur, Mson;
  float Ucur, Uson;
  float UUcur, UUson;
  float Vcur, Vson;
  configuration sonq,deltaq;
  fpoint3D curp[MAXNBP], sonp[MAXNBP];
  ipoint3D curld[MAXLEAD],sonld[MAXLEAD];
  double erand48();
  
  FK(curq,curp);
  Vcur = HIT(curp,V);
  
  trans_lead(curld);
  POTENTIAL(curld,heurpot,Mcur);
  Ucur = UUcur = Mcur;
  found  = 0;
  *length = 0;
  assign(curq,result[*length]);
  (*length)++;

  while(1){

    /* Termination condition */
    if(Mcur <= (2.0*therobot.nblead) ) found = 1;
    if(found) {
      *Uend = Ucur;
      return(found);
    }

    nbtrials = 0;
    UUson = MAXFLOAT;
    
    while(nbtrials < NBTMAX){
      nbtrials++;

      for(i=0;i<therobot.dof;++i){
	deltaq[i] = therobot.coefc[i]*((random()%3)-1);
	sonq[i] = curq[i] + deltaq[i];
      }

      for(i=3;i<therobot.dof;++i){
	if(sonq[i] < 0.0) sonq[i] += 1.0;
	if(sonq[i] >= 1.0) sonq[i] -= 1.0;
      }


      FK(sonq,sonp);

      /* extra contraints put on the robot */
      goodmove=goodson(sonq,sonp);
      if(!goodmove) continue;

      outside(sonp,out);
      if(out) continue;
      
      Vson=HIT(sonp,V);
      if(Vson) continue;
      else nb_no_hit++;

      trans_lead(sonld);

      POTENTIAL(sonld,heurpot,Mson);
      Uson = UUson = Mson;
      
      if(UUson >= UUcur) {
	continue;
      }
      else break;
      
    }

    /* We reached a local minimum */    
    if(nbtrials >= NBTMAX) {
      *Uend = Ucur;
      break;
    } 
    
    assign(sonq,curq); /* We are still going downhill */

    Ucur = Uson;
    UUcur = UUson;
    Mcur =  Mson;
    
    assign(curq,result[*length]);
    (*length)++;
    
  }/* END WHILE */
  
  return(found);
}

/*****************************************************************************
  function to find down motion
*****************************************************************************/

boolean brownian(configuration minlocq,float *Uminloc,configuration *result,
		 int *length,int *nbsteps,image V,image *heurpot){

  boolean goodmove,out,finish=TRUE;
  int i,nbtrials;
  float Ucur,Uson,Vcur,Vson;
  configuration curq,sonq,deltaq;
  fpoint3D curp[MAXNBP],sonp[MAXNBP];
  ipoint3D curld[MAXLEAD],sonld[MAXLEAD];
  
  assign(minlocq,curq);
  Ucur = *Uminloc;
  FK(curq,curp);
  Vcur = HIT(curp,V);
  trans_lead(curld);
  POTENTIAL(curld,heurpot,Ucur);
  *length  = 0;
  assign(minlocq,result[*length]);
  (*length)++;
  nbtrials = 0;
  
  while(nbtrials < *nbsteps){
    for(i=0;i<therobot.dof;++i){
	deltaq[i] = therobot.coefc[i]*((random()%3)-1);
	sonq[i] = curq[i] + deltaq[i];
    }
    
    for(i=3;i<therobot.dof;++i){
      if(sonq[i] < 0) sonq[i] += 1.0;
      if(sonq[i] >= 1.0) sonq[i] -= 1.0;
    }
    
    FK(sonq,sonp);

    goodmove=goodson(sonq,sonp);
    if(!goodmove) continue;
    
    outside(sonp,out);
    if(out) continue;
    
    Vson=HIT(sonp,V);
    if(Vson) continue;
    else nb_no_hit++;
    
    trans_lead(sonld);
    POTENTIAL(sonld,heurpot,Uson);
    
    nbtrials++;
    assign(sonq,curq); /* We are still going uphill */
    assign_lead(sonld,curld);
    
    assign(curq,result[*length]);
    (*length)++;
    
    if(Uson < *Uminloc) {
      finish=FALSE;
      break;
    }

  }/* END WHILE */
  *Uminloc = Uson;
  assign(curq,minlocq);
  *nbsteps=nbtrials;
  return(finish);

}






