/*****************************************************************************

  3D Randomized Motion Planner

  COPYRIGHT (c) 1990 
  THE BOARD OF TRUSTEES OF THE LELAND STANFORD JUNIOR UNIVERSITY
  ALL RIGHTS RESERVED

  No duplications, whole or partial, manual or electronic, may be made 
  without the express permission of the author. Any such copies, or 
  revisions thereof, must display this notice unaltered.

  Robotics Laboratory,
  Computer Science Department,
  Stanford University

  Original author: Jerome Barraquand
  Modified by Tsai-Yen Li (li@flamingo.Stanford.EDU)
  Advisor: Jean-Claude Latombe

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "defs.h"
#include "solidstruct.h"
#include "macro.h"
#include "solidp.h"
#define HALTING_PROBLEM 256

/***************************************************************************
  declare external variables
****************************************************************************/

extern robot therobot;
extern descript ws_des;
extern FILE *log_fd;
extern long random();
extern void bcopy();

/***************************************************************************
  smooth4: multigrid resolution - sample progressively higher frequencies,
  systematically.
****************************************************************************/

int smooth4(image V,int intervals,path *pX) {

  int path_len, prev_path_len, pt, radius, delta;
  int intervals2;
  config_node *head;
  int pt1, pt2;
  double factor;
  
  fprintf(stderr,"Smooth the path using method #4\n");
  factor=1.0;
  prev_path_len=0;
  path_len=pX->dim;
  /* outer loop */
  do {
    /* inner loop */
    delta=intervals/6;
    pt=delta;
    do {
      radius=(int)((factor*path_len)/2);
      pt1=MAX(0,pt-radius);
      pt2=MIN(path_len-1,pt+radius);
      intervals2=((pt2-pt1)/(2.0 * radius)) * intervals;
      prev_path_len=path_len;
      if ((head=collision_free(V,intervals2,pX,pt1,pt2,NULL))!=NULL) {
	path_len=shift_path(pX,head,pt1,pt2);
	if (prev_path_len != path_len)
	  printf("prev_path_len=%d path_len=%d\n", prev_path_len, path_len);
      };
      pt += delta;
    } while (pt < path_len);
    factor /= 2.0;
    intervals /= 2;
  } while (delta > 2);
  return(TRUE);
}

/***************************************************************************
  smooth3: multigrid resolution - sample progressively higher frequencies,
  randomly
****************************************************************************/

void smooth3(image V,int intervals,path *pX) {

  int path_len, prev_path_len, random_pt, radius, same_count, count_limit;
  int intervals2;
  config_node *head;
  int pt1, pt2;
  double factor;
  
  printf("Smooth the path using method #3\n");
  factor=1.0;
  prev_path_len=0;
  path_len=pX->dim;
  count_limit=10;

  /* outer loop */
  do {
    same_count=0;
    /* inner loop */
    do {
      random_pt=random() % path_len;
      radius=(int)((factor*path_len)/2);
      pt1=MAX(0,random_pt-radius);
      pt2=MIN(path_len-1,random_pt+radius);
      intervals2=((pt2-pt1)/(2.0 * radius)) * intervals;
      prev_path_len=path_len;
      if ((head=collision_free(V,intervals2,pX,pt1,pt2,NULL))!=NULL) {
	path_len=shift_path(pX,head,pt1,pt2);
	printf("prev_path_len=%d path_len=%d\n", prev_path_len, path_len);
      };
      same_count=((prev_path_len==path_len) ? (same_count+1) : 0);
    } while (same_count < count_limit);
    factor /= 2.0;
    intervals /= 2;
    count_limit *= 2;
  } while (intervals > 2);
}


/***************************************************************************
  smooth2: random sampling of configuration space.
****************************************************************************/

void smooth2(image V,int intervals,path *pX) {

  int path_len, prev_path_len;
  config_node *head;
  int point1, point2;

  prev_path_len=0;
  path_len=pX->dim;
  do {
    printf("prev_path_len=%d path_len=%d\n", prev_path_len, path_len);
    if ((head=smooth2_aux(V,intervals,pX,&point1,&point2))!=NULL) {
      prev_path_len=path_len;
      if (list_length(head)>(point2-point1)) {
	printf("interpolated list too long between %d and %d\n",
	       point1, point2);
	exit(-1);
      }
      path_len=shift_path(pX,head,point1,point2);
      /* write_smooth(new_filename(j++), pX); */
    };
  } while (prev_path_len != path_len);
}

/***************************************************************************
  smooth2_aux
****************************************************************************/

config_node *smooth2_aux(image V,int intervals,path *pX,
			 int *ppoint1,int *ppoint2) {

  int point1, point2, path_len,intervals2;
  config_node *head;
  double ratio;

  path_len=pX->dim;
  /* choose two different random points */
  do {
    point1=random() % path_len;
    point2=random() % path_len;
  } while (point1==point2);
  if (point1>point2)
    swap_int(&point1,&point2);
  /* invariant: point1<point2 */
  ratio=
    sqrt(sqr_euclidean_distance(pX->d[point1],pX->d[point2]))/(double)DEGREE;
  intervals2=ratio*intervals;
  head=collision_free(V,intervals2,pX,point1,point2,NULL);
  if (head!=NULL)
    printf("collision free...\n");
  *ppoint1=point1;
  *ppoint2=point2;
  return(head);
}

/***************************************************************************
  smooth1: recursive midpoint (or random) sampling of configuration space
****************************************************************************/

config_node *smooth1(image V,int intervals,path *pX,int start,int end,
		     config_node *head){
  int mid,guess,range;
  config_node *item, *new_head, *head1, *head2, *local_head;
  double ratio;

  if ((end-start) < 2) {
    printf("can't find smoothed path: start=%d end=%d\n", start, end);
    local_head=NULL;
    while (start<=end) {
      CONS(config_node,item,local_head);
      bcopy(pX->d[start++], item->q, sizeof(configuration));
    }
    return(local_head);
  }

  if ((new_head=collision_free(V,intervals,pX,start,end,NULL))!=NULL) {
    printf("straight line between: %d and %d is collision-free.\n",
	   start, end);
    return(new_head);
  }
  else {
    printf("straight line between: %d and %d is *NOT* collision-free.\n",
	   start, end);
    /* take random interval - turns out that taking just the mid-point is
     *  good enough, i.e. mid=(start+end)/2
     */
    range=end-start;
    guess=random() % range;
    ratio=(double)guess/(double)range;
    mid=start+guess;
    /* start to mid */
    printf("calling smooth: between start=%d and mid=%d\n", start, mid);
    head1=smooth1(V,(int)(ratio*intervals),pX,start,mid,head);
    /* mid to end */
    printf("calling smooth: between mid+1=%d and end=%d\n", mid+1, end);
    head2=smooth1(V,(int)((1.0-ratio)*intervals),pX,mid+1,end,head1);
    head2=append_config_node(head2,head1);
    return(head2);
  }
}


/***************************************************************************
  function to check if the given segment of path is coliision free 
  and legal configuration or not 
****************************************************************************/

config_node *collision_free(image V,int intervals,path *pX,int start, 
			    int end,config_node *head) {

  int bool;
  configuration config;
  configuration START,END;
  fpoint3D kine_config[MAXNBP];
  int i,j;
  config_node *item;

  for(i=0;i<DEGREE;++i){
    if (therobot.rot[i]) {
      if ((fabs((pX->d[start][i])-(pX->d[end][i])) > 0.5)) {
	if (pX->d[start][i] < 0.5) {
	  START[i] = pX->d[start][i]+1.0;
	  END[i] = pX->d[end][i];
	} else {
	  START[i] = pX->d[start][i];
	  END[i] = pX->d[end][i]+1.0;
	}
      }
      else {
	START[i] = pX->d[start][i];
	END[i] = pX->d[end][i];
      }
    } else {
      START[i] = pX->d[start][i];
      END[i] = pX->d[end][i];
    } 
  }

  for (i=1; i<intervals; i++) {
    INTERPOLATE(START, END, i, intervals, config);
    for (j=0;j< DEGREE;j++) {
      if (therobot.rot[j]) {
	if (config[j] >= 1.0) 
	  config[j] -= 1.0;
      }
    }

    CFK(config,kine_config);

    bool= goodson(config,kine_config);
    if (!bool) {
      free_config_node(head);
      return(NULL);
    }

    outside(kine_config,bool);
    if (bool) {
      free_config_node(head);
      return(NULL);
    }
    /* collision avoidance added */

    bool=HIT(kine_config,V);
    if (bool) {
      free_config_node(head);
      return(NULL);
    }
    CONS(config_node,item,head);
    bcopy(config,item->q,sizeof(configuration));
  }
  return(head);
}

/***************************************************************************
  Function to free the configuration list
***************************************************************************/

void free_config_node(config_node *ptr) {

  config_node *item,*tmp;
  
  item=ptr;
  while (item!=NULL) {
    tmp=item;
    item=item->next;
    free(tmp);
  }
}
    
/***************************************************************************
  Function to append a configuration to a list
***************************************************************************/

config_node *append_config_node(config_node *head1,config_node *head2) {

  config_node *item, *new_item;
  
  if (head1==NULL)
    return(head2);
  else {
    item=append_config_node(head1->next, head2);
    new_item=(config_node *)calloc(1,sizeof(config_node));
    bcopy(head1->q,new_item->q,sizeof(configuration));
    new_item->next=item;
    free(head1);
    return(new_item);
  }
}

/***************************************************************************
  Function to print a list of configuration
***************************************************************************/

void print_config_node(config_node *head) {

  config_node *item;
  int i=0;
  
  item=head;
  while ((i<=HALTING_PROBLEM) && (item!=NULL)) {
    printf("%d: %4.2f\n", i++,item->q[0]);
    item=item->next;
  }
}


/***************************************************************************
  Function to compute the length of a list
***************************************************************************/

int list_length(config_node *head) {

  config_node *item;
  int i;
  
  item=head;
  i=0;
  while (item!=NULL) {
    i++;
    item=item->next;
  }
  return(i);
}

/***************************************************************************
  write out path in reverse order, as it was consed up in reverse.
***************************************************************************/

void write_path(config_node *item, FILE *fout) {

  if (item==NULL)
    return;
  else {
    write_path(item->next, fout);
    fwrite(item->q,sizeof(configuration),1,fout);
  }
}

/***************************************************************************
  Function to swap two integers
***************************************************************************/

void swap_int(int *p1,int *p2) {

  static int temp;
  
  temp = *p1;
  *p1 = *p2;
  *p2 = temp;
}

/***************************************************************************
  Function to shift the position of a list of configurations
***************************************************************************/

int shift_path(path *pX, config_node *head, int point1, int point2) {

  int i, j, path_len, rest, skip, list_len, k;
  config_node *item;
  
  path_len=pX->dim;
  i=list_len=list_length(head);

  /* now shift rest of configuration array over - including end point */
  rest=path_len-point2;
  skip=(point2-point1)-list_len-1;
  if (skip) {
    for (j=0, k=rest-1; j<rest; j++, k--) {
      if (skip>0)
	bcopy(pX->d[point2+j], pX->d[point2-skip+j], sizeof(configuration));
      else {
	bcopy(pX->d[point2+k], pX->d[point2-skip+k], sizeof(configuration));
      }
    }
  }

  /* copy in reverse list of collision free points to array */
  item=head;
  while (item!=NULL) {
    k=point1+i;
    bcopy(item->q, pX->d[k], sizeof(configuration));
    i--;
    item=item->next;
  }
  /* set new path dim */
  pX->dim=(point1+1)+list_len+rest;
  return(pX->dim);
}

/***************************************************************************
  Compute the square root of euclidean distance
***************************************************************************/

double sqr_euclidean_distance(configuration q1, configuration q2) {

  double sum=0, delta;
  int i;

  for (i=0; i<3; i++) {
    delta=q2[i]-q1[i];
    sum += delta*delta;
  }
  for (i=3; i<DEGREE; i++) {
    delta=fabs(q2[i]-q1[i]);
    if (delta > 0.5) {
      delta=1.0-delta;
    }
    sum += delta*delta;
  }

  return(sum);
}

/***************************************************************************
  Print the smoothed path to a file
***************************************************************************/

void write_smooth(char *ismooth,path *pX) {

  FILE *fsmooth;
  descript nsmooth;

  if (!(fsmooth = fopen(ismooth,"w"))) {
    fprintf(stderr,"Cannot open smooth %s\n",ismooth);
    exit(-1);
  }

  printf("writing output file: %s - path_len=%d\n", ismooth, pX->dim);
  nsmooth.dimx=nsmooth.dimy=1;
  nsmooth.octet = sizeof(configuration);

  /* write the output path */
  nsmooth.dimz=pX->dim;
  fwrite(&nsmooth,sizeof(descript),1,fsmooth);
  fwrite(pX->d,nsmooth.octet,nsmooth.dimz,fsmooth);
  fclose(fsmooth);
}

/***************************************************************************
  Create a new file name 
***************************************************************************/

char *new_filename(int j) {

  static char name[]="00";
  
  name[0]=(j/10) + 48;
  name[1]=(j % 10) + 48;
  return(name);
}

/***************************************************************************
  Smooth the path so that the motion in one time step would be more even
***************************************************************************/

double curvilinear(path *X) {

  register int i, j;
  double length=0;
  double offset,difference;
  configuration *Xold;
  double *lg;
  int posit;

  printf("computing curvilinear approximation...\n");
  Xold = (configuration *)calloc(X->dim,sizeof(configuration));
  lg = (double *)calloc(X->dim,sizeof(double));

  /* save old path */
  bcopy(X->d[0], Xold[0], X->dim*sizeof(configuration));

  /* compute partial lengths of old path */
  for (i=0; i<X->dim-1; i++) {
    lg[i] = length;
    length += sqrt(sqr_euclidean_distance(Xold[i],Xold[i+1]));
  }
  lg[X->dim-1] = length;

  posit = 0;
  for (i=1; i<X->dim-1; i++) {
    while ((offset = (length*i/(double)(X->dim-1)-lg[posit+1])) > 0)
      posit++;
    
    for (j=0; j<3; j++) {
      X->d[i][j] = Xold[posit+1][j] +
	offset/(lg[posit+1]-lg[posit]) * (Xold[posit+1][j] - Xold[posit][j]);
    }

    for (j=3; j<DEGREE; j++) {
      difference = (Xold[posit+1][j] - Xold[posit][j]);      
      if (fabs(difference) > 0.5) {
	if (difference > 0.0) 
	  difference = 1.0 - difference;
	else
	  difference = 1.0 + difference;
      }
      X->d[i][j] = Xold[posit+1][j] +
	offset/(lg[posit+1]-lg[posit]) * difference;
      if (X->d[i][j] >= 1.0) 
	X->d[i][j] -= 1.0;
      if (X->d[i][j] < 0.0)
	X->d[i][j] += 1.0;
    }
  }
  free(Xold);
  free(lg);
  return(length);
}


/***************************************************************************
  Smooth the path so that the motion in one time step would be more even
***************************************************************************/

boolean CurvePath(configuration *X,int dim,image V) {

  register int i,j,k;
  double length=0;
  double offset,difference;
  configuration *Xold;
  double *lg;
  int posit;
  fpoint3D kine_conf[MAXNBP];

  if (dim<3) return(TRUE);

  Xold = (configuration *)calloc(dim,sizeof(configuration));
  lg = (double *)calloc(dim,sizeof(double));

  /* save old path */
  bcopy(X[0], Xold[0],dim*sizeof(configuration));
  /* compute partial lengths of old path */
  for (i=0; i<dim-1; i++) {
    lg[i] = length;
    length += sqrt(sqr_euclidean_distance(Xold[i],Xold[i+1]));
  }
  lg[dim-1] = length;
  
  posit = 0;
  for (i=0; i<dim-1; i++) {
    while ((offset = (length*i/(double)(dim-1)-lg[posit+1])) > 0)
      posit++;
    
    for (j=0; j<DEGREE; j++) {
      if (therobot.rot[j]) {
	difference = (Xold[posit+1][j] - Xold[posit][j]);      
	if (fabs(difference) > 0.5) {
	  if (difference > 0.0) 
	    difference = 1.0 - difference;
	  else
	    difference = 1.0 + difference;
	}
	
	X[i][j] = Xold[posit+1][j] +
	  offset/(lg[posit+1]-lg[posit]) * difference;
	if (X[i][j] >= 1.0) 
	  X[i][j] -= 1.0;
	if (X[i][j] < 0.0)
	  X[i][j] += 1.0;
      } else {
	X[i][j] = Xold[posit+1][j] +
	  offset/(lg[posit+1]-lg[posit]) * (Xold[posit+1][j] - Xold[posit][j]);
      }
    }

    CFK(X[i],kine_conf);
    if (HIT(kine_conf,V)) {
      bcopy(Xold[0],X[0],dim*sizeof(configuration));
      k=dim/2;
      free(Xold);
      free(lg);
      CurvePath((configuration *)&(X[0][0]),k,V);
      CurvePath((configuration *)&(X[k][0]),dim-k,V);
      return(FALSE);
    }
  }
  free(Xold);
  free(lg);
#ifdef DEBUG
  fprintf(log_fd,"curvilinear succeeds for length:%d\n",dim);
#endif
#ifdef VERBOSE
  if (verbose) 
    fprintf(stderr,"curvilinear succeeds for length:%d\n",dim);
#endif
  return(TRUE);
}

