/*****************************************************************************

  3D Randomized Motion Planner

  COPYRIGHT (c) 1990 
  THE BOARD OF TRUSTEES OF THE LELAND STANFORD JUNIOR UNIVERSITY
  ALL RIGHTS RESERVED

  No duplications, whole or partial, manual or electronic, may be made 
  without the express permission of the author. Any such copies, or 
  revisions thereof, must display this notice unaltered.

  Robotics Laboratory,
  Computer Science Department,
  Stanford University

  Original author: Jerome Barraqua
  Modified by Tsai-Yen Li (li@flamingo.Stanford.EDU)
  Advisor: Jean-Claude Latombe

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <time.h>
#include <memory.h>
#include "defs.h"
#include "solidstruct.h"
#include "macro.h"
#include "solidp.h"
#include "parse.h"

/***************************************************************************
  external variables
***************************************************************************/

descript ws_des,path_des;
robot therobot;
FILE *log_fd;
int verbose;
long nb_fk_calls=0;
long nb_hit_calls=0;
long nb_goodson_calls=0;

/***************************************************************************
  Function to initialize the robot parameters
***************************************************************************/

void initparams(FILE *fd_robot){

  fprintf(stderr,"Initialize parameters ... \n");
  Read_Robot_File(fd_robot);

}

/***************************************************************************
  Main function of smooth 
***************************************************************************/

int main(int argc, char **argv) {
  
  char bmap_name[80]; /* File of the 3d segments */
  char robot_name[80];
  char inpath_name[80]; /* Path input */
  char outpath_name[80]; /* Path output */
  FILE *fd_bmap,*fd_robot,*fd_inpath,*fd_outpath;
  register int i;
  int n,curvline=0;
  image V;
  path X;
  struct arg *args;
 
  args=parse_args(argc,argv);

  if (args[chindex('h')].ch) {
    PrintHelp();
    exit(0);
  }

  if (args[chindex('v')].ch) 
    verbose=1;

  if (args[chindex('c')].ch) 
    curvline=1;

  if (!GetStrOpt(args,'i',bmap_name,"bmap.dat",0,12)) 
    fprintf(stderr,"Use default bitmap file name %s\n",bmap_name);
  if (!GetStrOpt(args,'r',robot_name,"robot.dat",0,12)) 
    fprintf(stderr,"Use default robot data file name %s\n",robot_name);
  if (!GetStrOpt(args,'p',inpath_name,"path.dat",0,12)) 
    fprintf(stderr,"Use default input path file name %s\n",inpath_name);
  if (!GetStrOpt(args,'o',outpath_name,"paths.dat",0,12)) 
    fprintf(stderr,"Use default output path file name %s\n",outpath_name);
  if (!GetIntOpt(args,'n',&n,4,0,12)) 
    fprintf(stderr,"Use default number of smoothing: %d\n",n);

  if(!(fd_bmap = fopen(bmap_name,"r"))) {
    fprintf(stderr,"Cannot open file %s\n",bmap_name);	
    exit(-1);
  }
  
  if(!(fd_robot = fopen(robot_name,"r"))) {
    fprintf(stderr,"Cannot open file %s\n",robot_name);	
    exit(-1);
  }
  
  if(!(fd_inpath = fopen(inpath_name,"r"))) {
    fprintf(stderr,"Cannot open path %s\n",inpath_name);
    exit(-1);
  }

  if(!(fd_outpath = fopen(outpath_name,"w"))) {
    fprintf(stderr,"Cannot open path %s\n",outpath_name);
    exit(-1);
  }

  fread(&ws_des,sizeof(descript),1,fd_bmap);

  if((V.d = (unsigned char ***)calloc3D(ws_des.dimx,ws_des.dimy,ws_des.dimz,
					sizeof(unsigned char))) == NULL) {
    perror("Memory allocation error");
    exit(-1);
  }
  
  fread(&(V.d[0][0][0]),sizeof(unsigned char),ws_des.dimx*ws_des.dimy*ws_des.dimz,fd_bmap);
  
  fread(&path_des,sizeof(descript),1,fd_inpath);
  
  initparams(fd_robot);
  
  X.dim = path_des.dimz;
  X.d = (configuration *)calloc(X.dim,sizeof(configuration));
  fread(X.d,sizeof(configuration),X.dim,fd_inpath);
  
  for(i=0;i<n;i++) {
    smooth4(V,256,&X);
  }
#ifdef DEBUG
  for(i=0;i<X.dim;i++) {
    fprintf(log_fd,"i=%d,\t(%3.2f, %3.2f, %3.2f, %3.2f, %3.2f, %3.2f, %3.2f)\n",
	    i,X.d[i][0],X.d[i][1],X.d[i][2],X.d[i][3],X.d[i][4],X.d[i][5],X.d[i][6]);
  }
#endif
  if (curvline) {
    fprintf(stderr,"Polishing the path ...\n");
    CurvePath((configuration *)&(X.d[0][0]),X.dim,V);
    smooth4(V,256,&X);
#ifdef DEBUG      
    for(i=0;i<X.dim;i++) {
      fprintf(log_fd,"i=%d,\t(%3.2f, %3.2f, %3.2f, %3.2f, %3.2f, %3.2f, %3.2f)\n",
	      i,X.d[i][0],X.d[i][1],X.d[i][2],X.d[i][3],X.d[i][4],X.d[i][5],X.d[i][6]);
      CFK(X.d[i],qp);
      if (HIT(qp,V))
	fprintf(stderr,"Collision at step%d\n",i);
    }    
#endif      
  }
  
  path_des.octet = sizeof(configuration);
  path_des.dimx = 1;
  path_des.dimy = 1;
  path_des.dimz = X.dim;
  
  fwrite(&path_des,sizeof(descript),1,fd_outpath);  
  fwrite(X.d,path_des.octet,path_des.dimz,fd_outpath);

  fclose(fd_bmap);
  fclose(fd_robot);
  fclose(fd_inpath); 
  fclose(fd_outpath); 
  return(TRUE);
}


/***************************************************************************
  function to print help
***************************************************************************/

void PrintHelp(void) {

  fprintf(stderr,"Usage: \tsmooth [-i workfile][-r robotfile][-p inpathfile] [-o outpathfile][-n nb of smooth][-v][-c]\n");
  fprintf(stderr,"\t -i: workspace bitmap data file name [bmap.dat]\n");
  fprintf(stderr,"\t -p: input path file name [path.dat]\n");
  fprintf(stderr,"\t -o: output path file name [paths.dat]\n");
  fprintf(stderr,"\t -r: robot description file name [robot.dat]\n");
  fprintf(stderr,"\t -n: number of smoothing steps [0]\n");
  fprintf(stderr,"\t -c: polish path with curvilinear function\n");
  fprintf(stderr,"\t -v: verbose mode\n");

}
