/*****************************************************************************

  3D Randomized Motion Planner

  COPYRIGHT (c) 1990 
  THE BOARD OF TRUSTEES OF THE LELAND STANFORD JUNIOR UNIVERSITY
  ALL RIGHTS RESERVED

  No duplications, whole or partial, manual or electronic, may be made 
  without the express permission of the author. Any such copies, or 
  revisions thereof, must display this notice unaltered.

  Robotics Laboratory,
  Computer Science Department,
  Stanford University

  Original author: Jerome Barraquand
  Modified by Tsai-Yen Li (li@flamingo.Stanford.EDU)
  Advisor: Jean-Claude Latombe

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include "defs.h"
#include "solidstruct.h"
#include "macro.h"
#include "parse.h"
#include "synth3dp.h"

int verbose;
robot therobot;
descript ws_des;
FILE *log_fd;
long nb_fk_calls;

/*****************************************************************************
  Main function for synth3d
*****************************************************************************/

int main(int argc,char *argv[]) { 

  char bmap_name[80],seg_name[80],obj_name[80];   
  FILE *fd_bmap,*fd_seg,*fd_obj;
  int dimx,dimy,dimz;
  unsigned char ***ws_bmap;
  unsigned char ***V;
  unsigned char ***voro;
  segment *seg;
  object *world;
  int nbseg,nbobj;
  float grow;
  clock_t start,finish;
  struct arg *args;

  args=parse_args(argc,argv);

  if (args[chindex('h')].ch) {
    PrintHelp();
    exit(0);
  }

  if (args[chindex('v')].ch) {
    verbose=1;
  }

  GetIntOpt(args,'x',&dimx,128,32,256);
  GetIntOpt(args,'y',&dimy,128,32,256);
  GetIntOpt(args,'z',&dimz,128,32,256);
  fprintf(stderr,"Dimension of Work Space:dimx = %d dimy = %d dimz = %d\n",
	  dimx,dimy,dimz);

  if (!GetStrOpt(args,'d',obj_name,"object.dat",0,12)) 
    fprintf(stderr,"Use default object description file name:%s\n",obj_name);
  if (!GetStrOpt(args,'o',bmap_name,"bmap.dat",0,12)) 
    fprintf(stderr,"Use default workspace bitmap file name:%s\n",bmap_name);
    
  if (!(fd_obj = fopen(obj_name,"r"))) {
    fprintf(stderr,"Cannot open file %s\n",obj_name);
    exit(-1);
  }

  if (!(fd_bmap = fopen(bmap_name,"w"))) {
    fprintf(stderr,"Cannot open file %s\n",bmap_name);
    exit(-1);
  }

  strcpy(seg_name,bmap_name);
  strcat(seg_name,".seg");

  if(!(fd_seg = fopen(seg_name,"w"))) {
    fprintf(stderr,"Cannot open file %s\n",seg_name);
    exit(-1);
  }

  ws_des.octet = 1;
  ws_des.dimx = dimx;
  ws_des.dimy = dimy;
  ws_des.dimz = dimz;
  
  fwrite(&ws_des,sizeof(descript),1,fd_bmap);
  
  ws_bmap = (unsigned char ***)calloc3D(dimx,dimy,dimz,sizeof(unsigned char));
  V = (unsigned char ***)calloc3D(dimx,dimy,dimz,sizeof(unsigned char));
  voro = (unsigned char ***)calloc3D(dimx,dimy,dimz,sizeof(unsigned char));
  
  seg = (segment *) calloc(MAXSEG, sizeof(segment));

#ifdef DEBUG
  if((log_fd=fopen("debug.log","w"))==NULL) {
    fprintf(stderr,"Can't open debugging log file:debug.log\n");
    exit(-1);
  }
#endif

  Read_Object_File(fd_obj,&nbobj,&world,&grow);

  start=Start_Clock("synthese");
  Create_Bitmap(ws_des,ws_bmap,&nbseg,nbobj,seg,world);
  
  free(world);

  distance(ws_des,ws_bmap,V,voro,grow);
  finish=Finish_Clock("synthese");
  Calculate_Time("synthese",start,finish);

  start=Start_Clock("io");
  fwrite(&(V[0][0][0]),sizeof(unsigned char),dimx*dimy*dimz,fd_bmap);
  fwrite(&(voro[0][0][0]),sizeof(unsigned char),dimx*dimy*dimz,fd_bmap);
  fwrite(&nbseg,sizeof(int),1,fd_seg);
  fwrite(seg,sizeof(segment),nbseg,fd_seg);
  fclose(fd_bmap);
  fclose(fd_seg);
  finish=Finish_Clock("io");

  free(seg);
  free3D(ws_bmap);
  free3D(V);
  free3D(voro);
  
  Calculate_Time("io",start,finish);  
  return(1);
}

/**************************************************************************
  Function to read the object description file
***************************************************************************/

void Read_Object_File(FILE *fd_obj,int *nbobj,object **worldptr,float *grow) {

  register int i,j;

  fscanf(fd_obj,"%f\n",grow);
  fscanf(fd_obj,"%d\n",nbobj);

  (*worldptr) = (object *) calloc((*nbobj),sizeof(object));
  for (i=0;i<(*nbobj);i++) {
    fscanf(fd_obj,"%d\n",&((*worldptr)[i].nbprim));
    (*worldptr)[i].prim = 
      (primitive *) calloc((*worldptr)[i].nbprim,sizeof(primitive));
    for (j=0;j<(*worldptr)[i].nbprim;j++) {
      fscanf(fd_obj,"%d %d %f %f %f %f %f %f\n",
	     &((*worldptr)[i].prim[j].objtype),
	     &((*worldptr)[i].prim[j].optype),
	     &((*worldptr)[i].prim[j].orig[0]),
	     &((*worldptr)[i].prim[j].orig[1]),
	     &((*worldptr)[i].prim[j].orig[2]),
	     &((*worldptr)[i].prim[j].orig[3]),
	     &((*worldptr)[i].prim[j].orig[4]),
	     &((*worldptr)[i].prim[j].orig[5]));
      fscanf(fd_obj,"%lf %lf %lf\n",
	     &((*worldptr)[i].prim[j].dim.x),
	     &((*worldptr)[i].prim[j].dim.y),
	     &((*worldptr)[i].prim[j].dim.z));
#ifdef DEBUG      
      fprintf(log_fd,"%d %d %f %f %f %f %f %f\n",
	      ((*worldptr)[i].prim[j].objtype),
	      ((*worldptr)[i].prim[j].optype),
	      ((*worldptr)[i].prim[j].orig[0]),
	      ((*worldptr)[i].prim[j].orig[1]),
	      ((*worldptr)[i].prim[j].orig[2]),
	      ((*worldptr)[i].prim[j].orig[3]),
	      ((*worldptr)[i].prim[j].orig[4]),
	      ((*worldptr)[i].prim[j].orig[5]));
      fprintf(log_fd,"%f %f %f\n",
	     ((*worldptr)[i].prim[j].dim.x),
	      ((*worldptr)[i].prim[j].dim.y),
	     ((*worldptr)[i].prim[j].dim.z));
#endif
    }
  }
}

/***************************************************************************
  Function to print help
****************************************************************************/

void PrintHelp(void) {

  fprintf(stderr,"Usage: synth3d [-x dimx] [-y dimy] [-z dimz] [-d objectfile][-o outfile][-v]\n");
  fprintf(stderr,"\t -xyz: workspace dimension [128]\n");
  fprintf(stderr,"\t -o: workspace bitmap data file name [bmap.dat]\n");
  fprintf(stderr,"\t -d: workspace object description file name [robot.dat]\n");
  fprintf(stderr,"\t -v: verbose mode\n");
}

